from fabric.tasks import task
from xefab.utils import console


def in_path(c, command: str):
    result = c.run(f"which {command}", hide=True, warn=True)
    return result.ok


def ensure_dependency(dep, installer=None):
    @task
    def wrapper(c, *args, **kwargs):
        if in_path(c, dep):
            return
        if installer is not None:
            installer(c, *args, **kwargs)
        else:
            raise RuntimeError(f"Dependency {dep} not found.")
    return wrapper

@task(pre=[ensure_dependency('wget')])
def go(c, force: bool = False):
    if in_path(c, 'go') and not force:
        console.print('Go already installed on system.')
        return
    result = c.run('python -m platform', hide=True, warn=True)
    assert result.ok, "Failed to deduce system."
    system = result.stdout.split('-')[0]
    if system in ('Linux', 'Darwin'):
        c.run('wget -q -O - https://raw.githubusercontent.com/canha/golang-tools-install-script/master/goinstall.sh | bash')
        console.print('Go installed. Please add `export PATH=$PATH:/usr/local/go/bin` to your .bashrc/.profile file')
    else:
        raise RuntimeError(f"{system} currently not supported by this task.")


@task(pre=[ensure_dependency('go', installer=go)])
def gopass(c, force: bool = False):
    if in_path(c, 'gopass') and not force:
        console.print('Gopass already installed on system.')
        return
    c.run('go install github.com/gopasspw/gopass@latest')
    console.print('Done')


@task
def chezmoi(c, force: bool = False):
    if in_path(c, 'chezmoi') and not force:
        console.print('Chezmoi already installed on system.')
        return
    c.run('sh -c "$(curl -fsLS get.chezmoi.io)"')
    console.print('Done.')


@task
def github_cli(c, force: bool = False):
    if in_path(c, 'gh') and not force:
        console.print('Github CLI already installed on system.')
        return
    c.run('conda install gh --channel conda-forge')
    console.print('Done.')


@task(pre=[ensure_dependency('wget')])
def gnupg(c, force: bool = False):
    if in_path(c, 'gpg') and not force:
        console.print('GnuPG already installed on system.')
        return
    scripts = {
        'install_gpg_all.sh': 'https://raw.githubusercontent.com/rnpgp/gpg-build-scripts/master/install_gpg_all.sh',
        'install_gpg_component.sh': 'https://raw.githubusercontent.com/rnpgp/gpg-build-scripts/master/install_gpg_component.sh'
    }
        
    for name, script in scripts.items():
        c.run(f'wget -q {script}')
        c.run(f'chmod +x {name}')
    c.run('./install_gpg_all.sh')
    console.print('Done.')


import contextlib
import errno
import os
import socket
from typing import Optional
import time
from types import TracebackType
from typing import Optional, Type, Union, List

from rich.console import Console, RenderableType
from rich.jupyter import JupyterMixin
from rich.live import Live
from rich.spinner import Spinner
from rich.style import StyleType
from rich.panel import Panel
from rich.layout import Layout
from rich import box
import pandas as pd
from invoke.util import enable_logging
from rich.table import Table


console = Console()


if os.environ.get("XEFAB_DEBUG") in ("1", "true", "True"):
    enable_logging()


def get_open_port(start=5000, end=None, bind_address="", *socket_args, **socket_kwargs):
    if start < 1024:
        start = 1024

    if end is None:
        end = start + 10000
    port = start
    while port < end:
        try:
            with contextlib.closing(
                socket.socket(*socket_args, **socket_kwargs)
            ) as my_socket:
                my_socket.bind((bind_address, port))
                my_socket.listen(1)
                this_port = my_socket.getsockname()[1]
                return this_port
        except socket.error as error:
            if not error.errno == errno.EADDRINUSE:
                raise
        port += 1
    raise Exception("Could not find open port")


def df_to_table(
    pandas_dataframe: pd.DataFrame,
    rich_table: Table = None,
    show_index: bool = True,
    index_name: Optional[str] = None,
) -> Table:
    """Convert a pandas.DataFrame obj into a rich.Table obj.
    Args:
        pandas_dataframe (DataFrame): A Pandas DataFrame to be converted to a rich Table.
        rich_table (Table): A rich Table that should be populated by the DataFrame values.
        show_index (bool): Add a column with a row count to the table. Defaults to True.
        index_name (str, optional): The column name to give to the index column. Defaults to None, showing no value.
    Returns:
        Table: The rich Table instance passed, populated with the DataFrame values."""
    if rich_table is None:
        rich_table = Table(show_header=True, header_style="bold magenta")
    if show_index:
        index_name = str(index_name) if index_name else ""
        rich_table.add_column(index_name)

    for column in pandas_dataframe.columns:
        rich_table.add_column(str(column))

    for index, value_list in enumerate(pandas_dataframe.values.tolist()):
        row = [str(index)] if show_index else []
        row += [str(x) for x in value_list]
        rich_table.add_row(*row)

    return rich_table


class TaskLiveDisplay(JupyterMixin):
    
    def __init__(
        self,
        taskname: str,
        *,
        status: RenderableType = None,
        print_buffer: Union[str, List[str]] = None,
        log_buffer: Union[str, List[str]] = None,
        console: Optional[Console] = None,
        spinner: str = "dots",
        spinner_style: StyleType = "status.spinner",
        speed: float = 1.0,
        refresh_per_second: float = 12.5,
        transient: bool = False,
        show_log: bool = False,
        max_rows: int = 30, 
    ):
        
        self._taskname = taskname
        self._status = status
        self._log_buffer = log_buffer or []
        self._print_buffer = print_buffer or []
        self._show_log = show_log
        self._max_rows = max_rows
        self.spinner = spinner
        self.spinner_style = spinner_style
        self.speed = speed
        self._spinner = Spinner(spinner, text=self._status, style=self.spinner_style, speed=self.speed)
        
        self._live = Live(
            self,
            console=console,
            auto_refresh=True,
            refresh_per_second=refresh_per_second,
            transient=transient,
        )
    

    @property
    def show_log(self):
        return self._show_log
    
    @show_log.setter
    def show_log(self, value: bool):
        self._show_log = value
        
    def render(self):
        layout = Layout()
        layout.split_column(
            Layout(name="header"),
            Layout(name="output"),
            Layout(name="status"),
            )
        layout["output"].split_row(
            Layout(name="print"),
            Layout(name="log"),
            )
        layout.size = self._max_rows + 5
        layout["output"].size = self._max_rows
        layout["header"].size = 2
        layout["status"].size = 2
        layout["status"].visible = self._status is not None
        layout['output']['log'].visible = self.show_log

        header_panel = Panel("", subtitle=self._taskname, box=box.MINIMAL)
        print_panel = Panel('\n'.join(self._print_buffer)[:self._max_rows], title='Output', expand=True)
        log_panel = Panel('\n'.join(self._log_buffer)[:self._max_rows], title='Log', expand=True)
        
        layout['header'].update(header_panel)
        layout['output']['print'].update(print_panel)
        layout['output']['log'].update(log_panel)
        if self._status is not None:
            self._spinner.update(text=self._status)
            layout['status'].update(self._spinner)
        else:
            layout['status'].update("")
        return layout
    
    def status(self, value: RenderableType):
        self._status = value

    def log(self, *args):
        self._log_buffer.extend(map(str, args))
        
    def print(self, *args):
        self._print_buffer.extend(map(str, args))
        
    def start(self) -> None:
        """Start the status animation."""
        self._live.start()

    def stop(self) -> None:
        """Stop the spinner animation."""
        self._live.stop()

    def __rich__(self) -> RenderableType:
        return self.render()

    def __enter__(self) -> "Status":
        self.start()
        return self
    
    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        self.stop()




import time
import webbrowser
from io import BytesIO, StringIO
from random import choices
from string import ascii_lowercase

from fabric.tasks import task
from rich.layout import Layout
from rich.panel import Panel

from xefab.utils import console, get_open_port

from .squeue_task import parse_squeue_output
from .utils import print_splash


SPLASH_SCREEN = r"""
 __   __ ______  _   _   ____   _   _      _______ 
 \ \ / /|  ____|| \ | | / __ \ | \ | |    |__   __|
  \ V / | |__   |  \| || |  | ||  \| | _ __  | |   
   > <  |  __|  | . ` || |  | || . ` || '_ \ | |   
  / . \ | |____ | |\  || |__| || |\  || | | || |   
 /_/ \_\|______||_| \_| \____/ |_| \_||_| |_||_|   

                    The UChicago Analysis Center

"""

JOB_HEADER = """#!/bin/bash
#SBATCH --job-name=straxlab
#SBATCH --output={log_fn}
#SBATCH --error={log_fn}
#SBATCH --account=pi-lgrandi
#SBATCH --ntasks=1
#SBATCH --cpus-per-task={n_cpu}
#SBATCH --mem-per-cpu={mem_per_cpu}
#SBATCH --time={max_hours}:00:00
{extra_header}

export NUMEXPR_MAX_THREADS={n_cpu}
echo Starting jupyter job

"""

GPU_HEADER = """\
#SBATCH --partition=gpu2
#SBATCH --gres=gpu:1

module load cuda/10.1
"""

CPU_HEADER = """\
#SBATCH --qos {qos}
#SBATCH --partition {partition}
{reservation}
"""


# This is only if the user is NOT starting the singularity container
# (for singularity, starting jupyter is done in _xentenv_inner)
START_JUPYTER = """
JUP_PORT=$(( 15000 + (RANDOM %= 5000) ))
JUP_HOST=$(hostname -i)
echo $PYTHONPATH
jupyter {jupyter} --no-browser --port=$JUP_PORT --ip=$JUP_HOST --notebook-dir {notebook_dir} 2>&1
"""


SUCCESS_MESSAGE = """
All done! If you have linux, execute this command on your laptop:

ssh -fN -L {port}:{ip}:{port} {username}@{host} && sensible-browser http://localhost:{port}/{token}

If you have a mac, instead do:

ssh -fN -L {port}:{ip}:{port} {username}@{host} && open "http://localhost:{port}/{token}"

Happy strax analysis, {username}!
"""


START_NOTEBOOK_SH = """
#!/bin/bash

IMAGE_NAME=$1
JUPYTER_TYPE=$2
NOTEBOOK_DIR=$3

IMAGE_DIR='/project2/lgrandi/xenonnt/singularity-images'

# if we passed the full path to an image, use that
if [ -e ${IMAGE_NAME} ]; then
  CONTAINER=${IMAGE_NAME}
# otherwise check if the name exists in the standard image directory
elif [ -e ${IMAGE_DIR}/${IMAGE_NAME} ]; then
  CONTAINER=${IMAGE_DIR}/${IMAGE_NAME}
# if not any of those, throw an error
else
  echo "We could not find the container at its full path ${IMAGE_NAME} or in ${IMAGE_DIR}. Exiting."
  exit 1
fi

if [ "x${JUPYTER_TYPE}" = "x" ]; then
  JUPYTER_TYPE='lab'
fi

echo "Using singularity image: ${CONTAINER}"

PORT=$(( 15000 + (RANDOM %= 5000) ))
SINGULARITY_CACHEDIR=/scratch/midway2/$USER/singularity_cache

# script to run inside container
INNER=/home/$USER/straxlab/.singularity_inner_${PORT}.sh
cat > $INNER << EOF
#!/bin/bash
JUP_HOST=\$(hostname -i)
## print tunneling instructions
echo -e "
    Copy/Paste this in your local terminal to ssh tunnel with remote
    -----------------------------------------------------------------
    ssh -N -f -L localhost:$PORT:\$JUP_HOST:$PORT ${USER}@midway2.rcc.uchicago.edu
    -----------------------------------------------------------------

    Then open a browser on your local machine to the following address
    ------------------------------------------------------------------
    localhost:$PORT
    ------------------------------------------------------------------
    and use the token that appears below to login.

    OR replace "$ipnip" in the address below with "localhost" and copy
    to your local browser.
    " 2>&1

jupyter ${JUPYTER_TYPE} --no-browser --port=$PORT --ip=\$JUP_HOST --notebook-dir ${NOTEBOOK_DIR} 2>&1
EOF
chmod +x $INNER

module load singularity
singularity exec --bind /project2 --bind /scratch/midway2/$USER --bind /dali $CONTAINER $INNER

"""


@task(pre=[print_splash])
def start_jupyter(
    c,
    env: str = "singularity",
    partition: str = "xenon1t",
    bypass_reservation: bool = False,
    tag: str = "development",
    node: str = None,
    timeout: int = 120,
    cpu: int = 2,
    ram: int = 8000,
    gpu: bool = False,
    jupyter: str = "lab",
    local_cutax: bool = False,
    notebook_dir: str = None,
    max_hours: int = 4,
    force_new: bool = False,
    local_port=8889,
    detached: bool = False,
):
    """Start a jupyter notebook on remote host."""

    print = console.log

    REMOTE_HOME = f"/home/{c.user}"
    if notebook_dir is None:
        notebook_dir = REMOTE_HOME
    output_folder = f"{REMOTE_HOME}/straxlab"

    local_port = get_open_port(start=local_port)
    env_vars = {}

    if local_cutax:
        env_vars["INSTALL_CUTAX"] = "0"

    with console.status("Creating straxlab folder..."):
        c.run("mkdir -p " + output_folder)

    log_fn = output_folder + "/jupyter.log"

    if env == "singularity":
        s_container = "xenonnt-%s.simg" % tag
        starter_path = f"{output_folder}/start_notebook.sh"
        starter_script = StringIO(START_NOTEBOOK_SH)
        with console.status(f"Copying starter script to {c.host}:{starter_path} ..."):
            c.put(starter_script, remote=starter_path)
        with console.status("Making starter script executable..."):
            c.run(f"chmod +x {starter_path}")

        batch_job = (
            JOB_HEADER + "{starter_path} "
            "{s_container} {jupyter} {nbook_dir}".format(
                starter_path=starter_path,
                s_container=s_container,
                jupyter=jupyter,
                nbook_dir=notebook_dir,
            )
        )
    elif env == "cvmfs":
        batch_job = (
            JOB_HEADER
            + "source /cvmfs/xenon.opensciencegrid.org/releases/nT/%s/setup.sh" % (tag)
            + START_JUPYTER.format(jupyter=jupyter, notebook_dir=notebook_dir)
        )
        print("Using conda from cvmfs (%s) instead of singularity container." % (tag))

    elif env == "backup":
        if tag != "development":
            raise ValueError(
                "I'm going to give you the latest container, you cannot choose a version!"
            )
        batch_job = (
            JOB_HEADER
            + "source /dali/lgrandi/strax/miniconda3/bin/activate strax"
            + START_JUPYTER.format(jupyter=jupyter, notebook_dir=notebook_dir)
        )
        print("Using conda from cvmfs (%s) instead of singularity container." % (tag))

    if partition == "kicp":
        qos = "xenon1t-kicp"
    else:
        qos = partition

    url = None
    url_cache_fn = REMOTE_HOME + "/.last_jupyter_url"
    username = c.user

    _want_to_make_reservation = partition == "xenon1t" and (not bypass_reservation)
    if ram > 16000 and _want_to_make_reservation:
        print(
            "You asked for more than 16 GB total memory you cannot use the notebook "
            "reservation queue for this job! We will bypass the reservation."
        )

    if cpu >= 8 and _want_to_make_reservation:
        print(
            "You asked for more than 7 CPUs you cannot use the notebook reservation "
            "queue for this job! We will bypass the reservation."
        )
    use_reservation = (
        (not force_new) and _want_to_make_reservation and cpu < 8 and ram <= 16000
    )

    if use_reservation:
        with console.status(
            "Notebook reservation requested. Checking availability..."
        ):
            result = c.run("scontrol show reservations", hide=True, warn=True)
            if result.failed or "ReservationName=xenon_notebook" not in result.stdout:
                print("Notebook reservation does not exist, submitting a regular job.")
                use_reservation = False

    unique_id = "".join(choices(ascii_lowercase, k=6))
    with console.status("Checking for existing jobs..."):
        result = c.run(f"squeue -u {c.user} -n straxlab", hide=True, warn=True)
        df = parse_squeue_output(result.stdout)
        if len(df):
            job_fn = "/".join([output_folder, f"notebook_{unique_id}.sbatch"])
        else:
            job_fn = "/".join([output_folder, "notebook.sbatch"])
    if not force_new:
        log_fn = "/".join([output_folder, "notebook.log"])
    else:
        log_fn = "/".join([output_folder, f"notebook_forced_{unique_id}.log"])

    extra_header = (
        GPU_HEADER
        if gpu
        else CPU_HEADER.format(
            partition=partition,
            qos=qos,
            reservation=(
                "#SBATCH --reservation=xenon_notebook" if use_reservation else ""
            ),
        )
    )
    if node:
        extra_header += "\n#SBATCH --nodelist={node}".format(node=node)
    if max_hours is None:
        max_hours = 2 if gpu else 8
    else:
        max_hours = int(max_hours)

    batch_job = batch_job.format(
        log_fn=log_fn,
        max_hours=max_hours,
        extra_header=extra_header,
        n_cpu=cpu,
        mem_per_cpu=int(ram / cpu),
    )

    with console.status("Reseting log file..."):
        c.put(StringIO(""), remote=log_fn)

    with console.status("Copying batch job to remote host..."):
        c.put(StringIO(batch_job), remote=job_fn)

    with console.status("Setting permissions on batch job..."):
        c.run("chmod +x " + job_fn)

    with console.status("Submitting batch job..."):
        result = c.run("sbatch " + job_fn, env=env_vars, hide=True, warn=True)

    if result.failed:
        raise RuntimeError("Could not submit batch job. Error: " + result.stderr)

    job_id = int(result.stdout.split()[-1])
    print("Submitted job with ID: %d" % job_id)

    with console.status("Waiting for your job to start..."):
        for _ in range(timeout):
            result = c.run("squeue -j %d" % job_id, hide=True, warn=True)
            df = parse_squeue_output(result.stdout)
            if len(df) and df["ST"].iloc[0] == "R":
                break
            time.sleep(1)
        else:
            raise RuntimeError("Job did not start within %d seconds." % timeout)

    print("Job started.")

    with console.status("Waiting for jupyter server to start...") as status:
        url = None
        for _ in range(timeout):
            time.sleep(1)
            log_content = BytesIO()
            result = c.get(log_fn, local=log_content)
            log_content.seek(0)
            lines = [line.decode() for line in log_content.readlines()]

            for line in lines:
                if "http://" in line and "?token=" in line:
                    url = line.split()[-1]
                    break
            else:
                result = c.run("squeue -j %d" % job_id, hide=True, warn=True)
                df = parse_squeue_output(result.stdout)
                if not len(df):
                    raise RuntimeError(
                        "Job has exited."
                        "This can be because it was canceled or due to an internal error."
                    )
                continue
            break
        else:
            raise RuntimeError(
                "Jupyter server did not start within %d seconds." % timeout
            )

        print("\nJupyter started succesfully.")
        print(f"\t Remote URL: {url}")
        remote_host, remote_port = url.split("/")[2].split(":")
        if "token" in url:
            token = url.split("?")[1].split("=")[1]
        else:
            token = ""

    local_url = f"http://localhost:{local_port}?token={token}"
    msg = f"Forwarding remote address {remote_host}:{remote_port} to local port {local_port}..."
    with console.status(msg) as status:
        print(f"You can access the notebook at {local_url}\n")

        if detached:
            c.local(
                f"ssh -fN -L {local_port}:{remote_host}:{remote_port} {c.user}@{c.host} &",
                disown=True,
                hide=False,
                warn=True,
            )
            time.sleep(3)
            c.local(f"python -m webbrowser -t {local_url}", hide=True, warn=True)
        else:
            # Can never be too careful
            local_port = int(local_port)
            remote_port = int(remote_port)
            with c.forward_local(
                local_port, remote_port=remote_port, remote_host=remote_host
            ):
                status.update(
                    "Local port forwarding is active.\n"
                    "   Press ENTER or CTRL-C to deactivate and cancel job."
                )
                time.sleep(3)
                result = c.local(
                    f"python -m webbrowser -t {local_url}", hide=True, warn=True
                )
                try:
                    console.input()
                except KeyboardInterrupt:
                    print("Keyboard interrupt received.")
                except:
                    print("Unknown error.")
                finally:
                    status.update("Closing tunnel and canceling job...")
            try:
                c.run("scancel %d" % job_id, hide=True)
            except:
                print("Could not cancel job. Please cancel it manually.")
    print("Goodbye!")



from fabric.tasks import task

from xefab.utils import console


@task
def download_file(c, path: str, out: str = None, preserve_mode: bool = True):
    """Download a file from a remote server."""
    with console.status(f'Downloading {path} from {c.host}'):
        c.get(path, local=out, preserve_mode=preserve_mode)
    console.print(f"Done.")


@task
def upload_file(c, path: str, remote_path: str = None, preserve_mode: bool = True):
    """Upload a file to a remote server."""
    with console.status(f'Uploading {path} to {c.host}'):
        c.put(path, remote=remote_path, preserve_mode=preserve_mode)
    console.print(f"Done.")



"""Console script for xesites."""
import sys

from fabric.executor import Executor
from fabric.main import Fab
from invoke.collection import Collection
from invoke.util import debug

from xefab import __version__, tasks
from xefab.collection import XefabCollection
from xefab.config import Config


class XeFab(Fab):
    """XeSites CLI"""

    def parse_collection(self):
        if self.namespace is None:
            self.namespace = XefabCollection.from_module(tasks, name="root")
            self.namespace.load_objects_from_entry_points()

        if self.core.unparsed and self.core.unparsed[0] in self.namespace.collections:
            sitename = self.core.unparsed.pop(0)
            self.namespace = self.namespace.collections[sitename]
            hostnames = self.namespace._configuration.get("hostnames", None)
            debug(f"xefab: {sitename} hostnames: {hostnames}")
            self.config.configure_ssh_for_host(sitename, hostnames)
            if not self.args.hosts.value and hostnames is not None:
                self.args.hosts.value = sitename
        super().parse_collection()

    def list_tasks(self):
        super().list_tasks()

        if self.collection.name != "root":
            # We are not in the root collection.
            return

        tuples = []
        for name, collection in self.collection.collections.items():
            hostnames = collection.configuration().get("hostnames", [])
            if isinstance(hostnames, str):
                hostnames = hostnames.split(",")
            if hostnames:
                hostnames = ",".join(hostnames)
                tuples.append((name, hostnames))

        if tuples:
            print("\n Remote Hosts:\n")
            self.print_columns(tuples)


def make_program():
    return XeFab(
        name="xefab",
        version=__version__,
        executor_class=Executor,
        config_class=Config,
    )


program = make_program()


if __name__ == "__main__":
    sys.exit(program.run())  # pragma: no cover



import json
from io import StringIO, BytesIO
from fabric.tasks import task
from xefab.utils import console
import pandas as pd
from .install import ensure_dependency, github_cli
from .secrets import github_token


def github_api_call(c, path: str, 
                    method: str = "GET", 
                    raw: bool = False, 
                    hide: bool = False,
                    pages: int = 10,
                    per_page=50):
    promises = []
    for page in range(1, pages+1):
        cmd = f"gh api {path}?page={page}&per_page={per_page}"
        if method != "GET":
            cmd += f" -X {method}"
    
        promise = c.run(cmd, hide=True, warn=True, asynchronous=True)
        promises.append(promise)

    for promise in promises:
        result = promise.join()
        if result.failed:
            raise RuntimeError(f"Failed to get github keys: {result.stderr}")
        
        data = result.stdout
        if data == '[]':
            continue
        if raw:
            yield data
        else:
            yield from json.loads(data)


def add_recipient(c, key: dict, gnuhome: str = "~/.gnupg", gopass_store: str = "xenonnt"):
    import gnupg
    gpg = gnupg.GPG(homedir=gnuhome)
    gpg.import_keys(key['raw_key'])
    console.print(f"Added {key['key_id']} to {gnuhome}")
    result = c.run(f"gopass --yes recipients add --store {gopass_store} {key['key_id']}", hide=True, warn=True)
    if result.ok:
        console.print(f"Added {key['key_id']} to gopass")
    else:
        console.print(f"Failed to add {key['key_id']} to gopass: {result.stderr}")

@task(pre=[ensure_dependency('gh', installer=github_cli)])
def list_xenonnt_members(c, hide: bool = False):
    users = github_api_call(c, "orgs/XENONnT/members", hide=True)
    
    usernames = [user["login"] for user in users]
    if not hide:
        console.print(usernames)
        console.print(f'Found {len(usernames)} members.')
    return usernames


@task(pre=[ensure_dependency('gh', installer=github_cli)])
def list_xenon1t_members(c, hide: bool = False):
    users = github_api_call(c, "orgs/XENON1T/members", hide=True)
    
    usernames = [user["login"] for user in users]
    if not hide:
        console.print(usernames)
        console.print(f'Found {len(usernames)} members.')
    return usernames


@task(pre=[ensure_dependency('gh', installer=github_cli)])
def get_xenonnt_keys(c, hide: bool = False, add: bool = False):
    users = github_api_call(c, "orgs/XENONnT/members", hide=True)

    keys = []
    for user in users:
        username = user["login"]
        keys = github_api_call(c, f"users/{username}/gpg_keys", hide=True, pages=1)
        keys = [key for key in keys if key["can_encrypt_storage"] and key["raw_key"]]
        if not keys:
            continue
        keys.extend(keys)
        if not hide:
            console.print(f'Found {len(keys)} keys for {username}:')
            for key in keys:
                console.print(key["key_id"])
        if add:
            for key in keys:
                add_recipient(c, key)
    if add:
        c.run("gopass sync", warn=True)
    return keys


@task
def user_db(c, limit: int = None, hide: bool = False):
    users = c.config.xent_collection(collection='users').find({}, projection={'_id': 0})
    if limit is not None:
        users = users.limit(int(limit))
    df = pd.json_normalize(list(users))
    if not hide:
        console.print_json(df.to_json(orient='records'), indent=4)
    return df



from io import BytesIO, StringIO
import os
import tempfile
from typing import List
import uuid
from xefab.utils import console
from fabric.tasks import task
from fabric.connection import Connection


sbatch_template = """#!/bin/bash
#SBATCH --job-name={jobname}
#SBATCH --output={log}
#SBATCH --error={log}
#SBATCH --account=pi-lgrandi
#SBATCH --qos={qos}
#SBATCH --partition={partition}
#SBATCH --mem-per-cpu={mem_per_cpu}
#SBATCH --cpus-per-task={cpus_per_task}
{hours}
{job}
"""

SINGULARITY_DIR = '/project2/lgrandi/xenonnt/singularity-images'

def singularity_wrap(c: Connection, 
                    jobstring: str, tmpdir: str,
                    image: str, bind: List[str] = []):
    """Wraps a jobscript into another executable 
    file that can be passed to singularity exec"""

    if isinstance(bind, str):
        bind = [bind]

    bind = list(bind) + [tmpdir]

    filename = "xefabtmp_" + f"{uuid.uuid4()}.sh".replace('-', '')[-10:]
    
    filepath = '/'.join([tmpdir, filename])
    
    console.print(f'Using {filepath} on {c.host} for inner jobscript.')
    
    with console.status(f'Uploading inner jobscript to {filepath} on {c.host}'):
        fd = BytesIO(f'#!/bin/bash\n{jobstring}'.encode())  
        c.put(fd, remote=filepath)
        del fd
    
    with console.status(f'Changing {filepath} on {c.host} to executable'):
        c.run(f'chmod +x {filepath}')
    
    bind_string = " ".join([f"--bind {b}" for b in bind])
    image = '/'.join([SINGULARITY_DIR, image])

    new_job_string = f"""singularity exec {bind_string} {image} {filepath}
rm {filepath}
"""
    return new_job_string

@task
def submit_job(c: Connection,
               command: str,
               *,
               partition='xenon1t',
               qos='xenon1t',
               account='pi-lgrandi',
               jobname='xefab_job',
               dry_run=False,
               mem_per_cpu=1000,
               container='xenonnt-development.simg',
               bind=('/dali', '/project2'),
               cpus_per_task=1,
               hours=None,
               ):
    """
    Submit a job to the dali batch queue
    EXAMPLE
        from utilix import batchq
        import time
        job_log = 'job.log'
        batchq.submit_job('echo "say hi"', log=job_log)
        time.sleep(10) # Allow the job to run
        for line in open(job_log):
            print(line)
    :param jobstring: the command to execute
    :param partition: partition to submit the job to
    :param qos: qos to submit the job to
    :param account: account to submit the job to
    :param jobname: how to name this job
    :param dry_run: only print how the job looks like
    :param mem_per_cpu: mb requested for job
    :param container: name of the container to activate
    :param bind: which paths to add to the container
    :param cpus_per_task: cpus requested for job
    :param hours: max hours of a job
    :param kwargs: are ignored
    :return: None
    """
    console.print(f'Using {c.host} as host')
    console.print(f"job command: {command}")

    with console.status(f'Looking up SCRATCH directory on {c.host}'):
        result = c.run('echo $SCRATCH', hide=True, warn=True)
        if result.ok and result.stdout:
            TMPDIR = result.stdout.strip()
        else:
            TMPDIR = '.'

    TMPDIR = '/'.join([TMPDIR, 'tmp'])

    console.print(f'Using {TMPDIR} as temporary directory')

    with console.status(f'Creating temporary directory {TMPDIR} on {c.host}'):
        c.run(f'mkdir -p {TMPDIR}', hide=True, warn=True)
    
    console.print(f'Created temporary directory {TMPDIR} on {c.host}')

    random_string = f"{uuid.uuid4()}".replace('-', '')[-8:]

    batchname = f"{jobname}_{random_string}"

    if container:
        # need to wrap job into another executable
        jobstr = singularity_wrap(c, command, TMPDIR, container, bind)
        jobstr = 'unset X509_CERT_DIR CUTAX_LOCATION\n' + 'module load singularity\n' + jobstr
    else:
        jobstr = command

    if hours is not None:
        hours = '#SBATCH --time={:02d}:{:02d}:{:02d}'.format(int(hours), int(hours * 60 % 60), int(hours * 60 % 60 * 60 % 60))
    else:
        hours = ''

    sbatch_file = f"{batchname}.sbatch"
    log = f"{batchname}.log"

    sbatch_script = sbatch_template.format(jobname=jobname, log=log, qos=qos, partition=partition,
                                           account=account, job=jobstr, mem_per_cpu=mem_per_cpu,
                                           cpus_per_task=cpus_per_task, hours=hours)

    if dry_run:
        console.print("=== DRY RUN ===")
        console.print(sbatch_script)
        return        

    console.print(f'Using {sbatch_file} as sbatch filename')

    with console.status(f'Uploading outer jobscript to {sbatch_file} on {c.host}'):
        fd = BytesIO(sbatch_script.encode())
        c.put(fd, remote=sbatch_file)


    command = "sbatch %s" % sbatch_file
    with console.status(f'Executing {command} on {c.host}'):
        result = c.run(command, hide=True, warn=True)
        if result.ok and result.stdout:
            job_id = result.stdout.strip()
            console.print(f'Job ID: {job_id}')

    with console.status(f'Removing {sbatch_file} on {c.host}'):
        c.run(f'rm {sbatch_file}', hide=True, warn=True)




from fabric.tasks import task
from xefab.utils import console
from invoke.context import DataProxy


def printable(d):
    if isinstance(d, (dict, DataProxy)):
        return {k: printable(v) for k,v in d.items()}
    elif isinstance(d, type):
        return d.__qualname__
    return d


@task
def show_config(c, config_name: str = None):
    """Get a config from the config server."""
    if config_name is None:
        console.print_json(data=printable(c.config), indent=4)
        return
    result = getattr(c.config, config_name, None)
    if result is None:
        result = c.config.get(config_name, None)
    if result is None:
        console.print(f"Config {config_name} not found.")
    result = printable(result)
    if isinstance(result, dict):
        console.print_json(data=result, indent=4)
    else:
        console.print(f"{config_name}: {result}")



import datetime
import pandas as pd

from fabric.tasks import task
from xefab.collection import XefabCollection
from xefab.utils import console, df_to_table


namespace = XefabCollection("osg")

namespace.configure({"hostnames": "login.xenon.ci-connect.net"})

slots = {
    'OWNER': 1,
    'BATCH_NAME': 1, 
    'SUBMITTED': 2,
    'DONE': 1,
    'RUN': 1,
    'IDLE': 1,
    'HOLD': 1,
    'TOTAL': 1,
    'JOB_IDS': 3,
}

mergers = {
    "SUBMITTED": lambda x: pd.to_datetime(f"{datetime.datetime.utcnow().year}/{x[0]}T{x[1]}"),
    "JOB_IDS": lambda x: "".join(x)
}


def parse_row(line, columns):
    parts = line.split()
    row = {}
    for column in columns:
        value = [parts.pop(0) for _ in range(slots[column])]
        merger = mergers.get(column, lambda x: x[0])
        row[column] = merger(value)
        
    return row


def parse_condorq_output(condorq_output):
    lines = condorq_output.split('\n')
    rows = []
    columns = []
    for line in lines:
        if line.startswith('OWNER'):
            columns = line.split()
            continue


        if not columns:
            continue
            
        if not line:
            break

        row = parse_row(line, columns)
        rows.append(row)
    
    df = pd.DataFrame(rows, columns=columns).dropna(how="all")
    return df


@task(aliases=["job-queue"])
def condor_q(c, all: bool = False, hide: bool = False):
    cmd = "condor_q"
    if all:
        cmd += " -allusers"
    result = c.run(cmd, hide=True, warn=True)
    if result.failed:
        console.print(f"Remote execution of {cmd} on {c.host} failed. stderr: {result.stderr}")
    df = parse_condorq_output(result.stdout)
    if not hide:
        table = df_to_table(df)
        console.print(table)
    return df
namespace.add_task(condor_q)



from xefab.collection import XefabCollection

from .jupyter_task import start_jupyter
from .squeue_task import squeue
from .transfer_tasks import download_file, upload_file
from .batchq import submit_job

namespace = XefabCollection("dali")

namespace.configure(
    {"hostnames": ["dali-login2.rcc.uchicago.edu", "dali-login1.rcc.uchicago.edu"]}
)

namespace.add_task(squeue)
namespace.add_task(start_jupyter)
namespace.add_task(download_file)
namespace.add_task(upload_file)
namespace.add_task(submit_job)



from xefab.collection import XefabCollection

from . import install, secrets, admin, utils
namespace = XefabCollection.from_module(utils , 'utils')

install = XefabCollection.from_module(install, name='install')
namespace.add_collection(install)

secret = XefabCollection.from_module(secrets, name='secrets')
namespace.add_collection(secret)

admin = XefabCollection.from_module(admin, name='admin')
namespace.add_collection(admin)


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['tests', 'xefab', 'xefab.hosts', 'xefab.hosts.uchicago', 'xefab.tasks']

package_data = \
{'': ['*']}

install_requires = \
['appdirs>=1.4.4,<2.0.0',
 'fabric>=3.0.0,<4.0.0',
 'gnupg>=2.3.1,<3.0.0',
 'pandas',
 'pymongo',
 'rich>=13.3.1,<14.0.0']

entry_points = \
{'console_scripts': ['xefab = xefab.main:program.run'],
 'xefab.tasks': ['dali = xefab.hosts.uchicago.dali',
                 'midway = xefab.hosts.uchicago.midway',
                 'osg = xefab.hosts.osg']}

setup_kwargs = {
    'name': 'xefab',
    'version': '0.1.0',
    'description': 'Top-level package for xefab.',
    'long_description': "=====\nxefab\n=====\n\nFabric based task execution for the XENON dark matter experiment\n\n\nInstallation\n------------\n\nTo install xefab, its recomended to use pipx_:\n\n.. code-block:: console\n\n    $ pipx install xefab\n\nAlternatively you can install it using pip:\n\n.. code-block:: console\n\n    $ pip install xefab\n\nUsage\n-----\n\nYou can list the available tasks and options by running xefab without any options.\n\n.. code-block:: console\n\n    $ xefab\n    Usage: xefab [--core-opts] <subcommand> [--subcommand-opts] ...\n\n    Core options:\n\n    --complete                         Print tab-completion candidates for given parse remainder.\n    --hide=STRING                      Set default value of run()'s 'hide' kwarg.\n    --print-completion-script=STRING   Print the tab-completion script for your preferred shell (bash|zsh|fish).\n    --prompt-for-login-password        Request an upfront SSH-auth password prompt.\n    --prompt-for-passphrase            Request an upfront SSH key passphrase prompt.\n    --prompt-for-sudo-password         Prompt user at start of session for the sudo.password config value.\n    --write-pyc                        Enable creation of .pyc files.\n    -d, --debug                        Enable debug output.\n    -D INT, --list-depth=INT           When listing tasks, only show the first INT levels.\n    -e, --echo                         Echo executed commands before running.\n    -f STRING, --config=STRING         Runtime configuration file to use.\n    -F STRING, --list-format=STRING    Change the display format used when listing tasks. Should be one of: flat (default), nested,\n                                        json.\n    -h [STRING], --help[=STRING]       Show core or per-task help and exit.\n    -H STRING, --hosts=STRING          Comma-separated host name(s) to execute tasks against.\n    -i, --identity                     Path to runtime SSH identity (key) file. May be given multiple times.\n    -l [STRING], --list[=STRING]       List available tasks, optionally limited to a namespace.\n    -p, --pty                          Use a pty when executing shell commands.\n    -R, --dry                          Echo commands instead of running.\n    -S STRING, --ssh-config=STRING     Path to runtime SSH config file.\n    -t INT, --connect-timeout=INT      Specifies default connection timeout, in seconds.\n    -T INT, --command-timeout=INT      Specify a global command execution timeout, in seconds.\n    -V, --version                      Show version and exit.\n    -w, --warn-only                    Warn, instead of failing, when shell commands fail.\n\n    Subcommands:\n\n    show-config                        Get a config from the config server.\n    admin.get-xenonnt-keys\n    admin.github-cli\n    admin.github-token\n    admin.list-xenon1t-members\n    admin.list-xenonnt-members\n    admin.user-db\n    dali.download-file                 Download a file from a remote server.\n    dali.squeue (dali.job-queue)       Get the job-queue status.\n    dali.start-jupyter                 Start a jupyter notebook on remote host.\n    dali.submit-job                    Submit a job to the dali batch queue\n    dali.upload-file                   Upload a file to a remote server.\n    install.chezmoi\n    install.github-cli\n    install.gnupg\n    install.go\n    install.gopass\n    midway.download-file               Download a file from a remote server.\n    midway.squeue (midway.job-queue)   Get the job-queue status.\n    midway.start-jupyter               Start a jupyter notebook on remote host.\n    midway.submit-job                  Submit a job to the dali batch queue\n    midway.upload-file                 Upload a file to a remote server.\n    osg.condor-q (osg.job-queue)\n    secrets.github-token\n    secrets.setup\n    secrets.setup-utilix-config\n\n\n    Remote Hosts:\n\n    dali     dali-login2.rcc.uchicago.edu,dali-login1.rcc.uchicago.edu\n    midway   midway2.rcc.uchicago.edu,midway2-login1.rcc.uchicago.edu,midway2-login2.rcc.uchicago.edu\n    osg      login.xenon.ci-connect.net\n\nSome tasks are registered to run on a specific host. When you run them, the --hosts option will be ignored.\n\ne.g. if you run\n\n.. code-block:: console\n\n    $ xefab midway start-jupyter\n\nThe task will be run on the midway host, not the host you specified with --hosts.\n\n\nFeatures\n--------\n\n* TODO\n\nCredits\n-------\n\nThis package was created with Cookiecutter_ and the `briggySmalls/cookiecutter-pypackage`_ project template.\n\n.. _Cookiecutter: https://github.com/audreyr/cookiecutter\n.. _`briggySmalls/cookiecutter-pypackage`: https://github.com/briggySmalls/cookiecutter-pypackage\n.. _pipx: https://github.com/pypa/pipx\n",
    'author': 'Yossi Mosbacher',
    'author_email': 'joe.mosbacher@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/jmosbacher/xefab',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8',
}


setup(**setup_kwargs)
